from asyncio import sleep
import tomli
from pathlib import Path
from raftify import Peers, Raft


def load_peers(filename: str) -> Peers:
    path = Path(__file__).parent / "fixtures" / filename
    cfg = tomli.loads(path.read_text())["raft"]["peers"]

    return Peers(
        {int(entry["node_id"]): f"{entry['host']}:{entry['port']}" for entry in cfg}
    )


async def wait_for_until_cluster_size_increase(raft: Raft, target: int):
    print("Waiting for cluster size to increase to {}...", target)

    while True:
        size = await raft.get_raft_node().get_cluster_size()
        if size >= target:
            break

        await sleep(0.1)

    # Wait for the conf_change reflected to the cluster
    await sleep(0.1)


async def wait_for_until_cluster_size_decrease(raft: Raft, target: int):
    print("Waiting for cluster size to decrease to {}...", target)

    while True:
        size = await raft.get_raft_node().get_cluster_size()
        if size <= target:
            break

        await sleep(0.1)

    # Wait for the conf_change reflected to the cluster
    await sleep(0.1)



import asyncio
import pytest
from utils import load_peers, wait_for_until_cluster_size_increase
from constant import THREE_NODE_EXAMPLE
from harness.raft_server import RAFTS, handle_bootstrap, run_rafts, wait_for_termination
from harness.state_machine import SetCommand


@pytest.mark.asyncio
async def test_data_replication():
    peers = load_peers(THREE_NODE_EXAMPLE)
    asyncio.create_task(run_rafts(peers))
    asyncio.create_task(wait_for_termination())

    await asyncio.sleep(2)

    await handle_bootstrap(peers)
    await asyncio.sleep(2)

    raft_1 = RAFTS.get(1)
    await wait_for_until_cluster_size_increase(raft_1, 3)
    entry = SetCommand("1", "test")

    raft_node_1 = raft_1.get_raft_node()
    raft_node_1.prepare_proposal(entry.encode())
    await raft_node_1.propose()

    await asyncio.sleep(1)

    # Data should be replicated to all nodes.
    for raft in RAFTS.values():
        store = await raft.get_raft_node().store()
        assert store.get("1") == "test"

    for raft in RAFTS.values():
        await raft.get_raft_node().quit()



import asyncio
import pickle
from raftify import RaftServiceClient


class SetCommand:
    """
    Represent simple key-value command.
    Use pickle to serialize the data.
    """

    def __init__(self, key: str, value: str) -> None:
        self.key = key
        self.value = value

    def encode(self) -> bytes:
        return pickle.dumps(self.__dict__)

    @classmethod
    def decode(cls, packed: bytes) -> "SetCommand":
        unpacked = pickle.loads(packed)
        return cls(unpacked["key"], unpacked["value"])


async def main() -> None:
    """
    A simple set of commands to test and show usage of RaftServiceClient.
    Please bootstrap the Raft cluster before running this script.
    """

    client = await RaftServiceClient.build("127.0.0.1:60061")
    client.prepare_propose(SetCommand("1", "A").encode())
    await client.propose()


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
from contextlib import suppress
import pickle
import sys
from typing import Optional
from raftify import (
    cli_main,
    set_confchange_context_deserializer,
    set_confchangev2_context_deserializer,
    set_entry_context_deserializer,
    set_entry_data_deserializer,
    set_message_context_deserializer,
    set_snapshot_data_deserializer,
    set_fsm_deserializer,
    set_log_entry_deserializer,
)


class SetCommand:
    """
    Represent simple key-value command.
    Use pickle to serialize the data.
    """

    def __init__(self, key: str, value: str) -> None:
        self.key = key
        self.value = value

    def encode(self) -> bytes:
        return pickle.dumps(self.__dict__)

    @classmethod
    def decode(cls, packed: bytes) -> "SetCommand":
        unpacked = pickle.loads(packed)
        return cls(unpacked["key"], unpacked["value"])


def pickle_deserialize(data: bytes) -> str | None:
    if data == b"":
        return None

    if pickle.PROTO in data:
        r = pickle.loads(data[data.index(pickle.PROTO) :])
        return r

    # Not pickle data
    return None


def register_custom_deserializer() -> None:
    """
    Initialize the custom deserializers.
    """

    set_confchange_context_deserializer(pickle_deserialize)
    set_confchangev2_context_deserializer(pickle_deserialize)
    set_entry_context_deserializer(pickle_deserialize)
    set_entry_data_deserializer(pickle_deserialize)
    set_message_context_deserializer(pickle_deserialize)
    set_snapshot_data_deserializer(pickle_deserialize)
    set_fsm_deserializer(pickle_deserialize)
    set_log_entry_deserializer(pickle_deserialize)


class HashStore:
    """
    A simple key-value store that stores data in memory.
    Use pickle to serialize the data.
    """

    def __init__(self):
        self._store = dict()

    def get(self, key: str) -> Optional[str]:
        return self._store.get(key)

    def as_dict(self) -> dict:
        return self._store

    def apply(self, msg: bytes) -> bytes:
        message = SetCommand.decode(msg)
        self._store[message.key] = message.value
        return msg

    def snapshot(self) -> bytes:
        return pickle.dumps(self._store)

    def restore(self, snapshot: bytes) -> None:
        self._store = pickle.loads(snapshot)


def main():
    register_custom_deserializer()

    argv = sys.argv
    argv.pop(0)
    argv.insert(0, "raftify_cli")

    with suppress(KeyboardInterrupt):
        asyncio.run(cli_main(argv))



import abc
from dataclasses import dataclass
from typing import Any, Callable, Final, Optional

# TODO: Make these abstract types available in the Python side.
class AbstractLogEntry(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def encode(self) -> bytes:
        raise NotImplementedError
    @classmethod
    def decode(cls, packed: bytes) -> "AbstractLogEntry":
        raise NotImplementedError

class AbstractStateMachine(metaclass=abc.ABCMeta):
    """
    A Finite State Machine (FSM) class.
    This class is designed to apply commands to a state, take snapshots of the state,
    and restore the state from a snapshot.
    """

    @abc.abstractmethod
    def apply(self, message: bytes) -> bytes:
        raise NotImplementedError
    @abc.abstractmethod
    def snapshot(self) -> bytes:
        raise NotImplementedError
    @abc.abstractmethod
    def restore(self, snapshot: bytes) -> None:
        raise NotImplementedError
    @abc.abstractmethod
    def encode(self) -> bytes:
        raise NotImplementedError
    @classmethod
    def decode(cls, packed: bytes) -> "AbstractStateMachine":
        raise NotImplementedError

class ReadOnlyOption:
    """Determines the relative safety of and consistency of read only requests."""

    Safe: Final[Any]
    """
    Safe guarantees the linearizability of the read only request by
    communicating with the quorum. It is the default and suggested option.
    """

    LeaseBased: Final[Any]
    """
    LeaseBased ensures linearizability of the read only request by
    relying on the leader lease. It can be affected by clock drift.
    If the clock drift is unbounded, leader might keep the lease longer than it
    should (clock can move backward/pause without any bound). ReadIndex is not safe
    in that case.
    """

class Logger:
    """ """

    def __init__(self) -> None: ...
    @staticmethod
    def default() -> "Logger": ...
    @staticmethod
    def new_file_logger(
        log_path: str, chan_size: int, rotate_size: int, rotate_keep: int
    ): ...
    def info(self, s: str) -> None:
        """
        Log info level record

        See `slog_log` for documentation.
        """
    def debug(self, s: str) -> None:
        """
        Log debug level record

        See `slog_debug` for documentation.
        """
    def trace(self, s: str) -> None:
        """
        Log trace level record

        See `slog_trace` for documentation.
        """
    def crit(self, s: str) -> None:
        """
        Log crit level record

        See `slog_crit` for documentation.
        """
    def error(self, s: str) -> None:
        """
        Log error level record

        See `slog_error` for documentation.
        """

class Raft:
    def __init__(self) -> None:
        """ """
    @staticmethod
    def build(
        node_id: int,
        raft_addr: str,
        fsm: AbstractStateMachine,
        config: "Config",
        logger: "Logger",
        initial_peers: Optional["Peers"] = None,
    ) -> "Raft":
        """ """
    @staticmethod
    async def request_id(self, peer_addr: str) -> "ClusterJoinTicket":
        """"""
    def prepare_join(self, join_ticket: "ClusterJoinTicket") -> None:
        """ """
    async def join(self) -> None:
        """ """
    def prepare_member_bootstrap_ready(self, leader_addr: str, node_id: int) -> None:
        """ """
    async def member_bootstrap_ready(self) -> None:
        """ """
    async def run(self) -> None:
        """ """
    async def snapshot(self) -> None:
        """ """
    def is_finished(self) -> bool:
        """ """
    def get_raft_node(self) -> "RaftNode":
        """ """

class RaftNode:
    async def is_leader(self) -> bool:
        """ """
    async def get_id(self) -> int:
        """ """
    async def get_leader_id(self) -> int:
        """ """
    async def get_peers(self) -> "Peers":
        """ """
    def prepare_add_peer(self, id: int, addr: str) -> None:
        """ """
    async def add_peer(self) -> None:
        """ """
    async def inspect(self) -> str:
        """ """
    def prepare_proposal(self, message: bytes) -> None:
        """ """
    async def propose(self) -> None:
        """ """
    def prepare_change_config(self, conf_change: "ConfChangeV2") -> None:
        """ """
    async def change_config(self) -> None:
        """ """
    def prepare_message(self, message: "Message") -> None:
        """ """
    async def send_message(self) -> None:
        """ """
    async def leave(self) -> None:
        """ """
    async def quit(self) -> None:
        """ """
    async def get_cluster_size(self) -> int:
        """ """
    async def store(self) -> AbstractStateMachine:
        """ """

class ClusterJoinTicket:
    """ """

    def __init__(
        self, reserved_id: int, leader_id: int, leader_addr: str, peers: "Peers"
    ) -> None: ...
    def get_reserved_id(self) -> int:
        """ """

class Peers:
    """ """

    def __init__(self, peers: dict) -> None: ...
    def to_dict(self) -> dict[int, str]: ...
    def items(self) -> list[tuple[int, str]]: ...
    def get(self, node_id: int) -> str: ...
    def add_peer(self, node_id: int, addr: str) -> None: ...
    def remove(self, node_id: int) -> None: ...
    def get_node_id_by_addr(self, addr: str) -> int: ...

class RaftConfig:
    """
    Config contains the parameters to start a raft.
    """

    def __init__(
        self,
        *,
        id: Optional[int] = None,
        election_tick: Optional[int] = None,
        heartbeat_tick: Optional[int] = None,
        applied: Optional[int] = None,
        max_size_per_msg: Optional[int] = None,
        max_inflight_msgs: Optional[int] = None,
        check_quorum: Optional[bool] = None,
        pre_vote: Optional[bool] = None,
        min_election_tick: Optional[int] = None,
        max_election_tick: Optional[int] = None,
        read_only_option: Optional["ReadOnlyOption"] = None,
        skip_bcast_commit: Optional[bool] = None,
        batch_append: Optional[bool] = None,
        priority: Optional[int] = None,
        max_uncommitted_size: Optional[int] = None,
        max_committed_size_per_ready: Optional[int] = None,
    ) -> None:
        """
        :param id: The identity of the local raft. It cannot be 0, and must be unique in the group.

        :param election_tick: The number of node.tick invocations that must pass between
        elections. That is, if a follower does not receive any message from the
        leader of current term before ElectionTick has elapsed, it will become
        candidate and start an election. election_tick must be greater than
        HeartbeatTick. We suggest election_tick = 10 * HeartbeatTick to avoid
        unnecessary leader switching

        :param heartbeat_tick: HeartbeatTick is the number of node.tick invocations that must pass between
        heartbeats. That is, a leader sends heartbeat messages to maintain its
        leadership every heartbeat ticks.

        :param applied: Applied is the last applied index. It should only be set when restarting
        raft. raft will not return entries to the application smaller or equal to Applied.
        If Applied is unset when restarting, raft might return previous applied entries.
        This is a very application dependent configuration.

        :param max_size_per_msg: Limit the max size of each append message. Smaller value lowers
        the raft recovery cost(initial probing and message lost during normal operation).
        On the other side, it might affect the throughput during normal replication.
        Note: math.MaxUusize64 for unlimited, 0 for at most one entry per message.

        :param max_inflight_msgs: Limit the max number of in-flight append messages during optimistic
        replication phase. The application transportation layer usually has its own sending
        buffer over TCP/UDP. Set to avoid overflowing that sending buffer.

        :param check_quorum: Specify if the leader should check quorum activity. Leader steps down when
        quorum is not active for an electionTimeout.

        :param pre_vote: Enables the Pre-Vote algorithm described in raft thesis section
        9.6. This prevents disruption when a node that has been partitioned away
        rejoins the cluster.

        :param min_election_tick: The range of election timeout. In some cases, we hope some nodes has less possibility
        to become leader. This configuration ensures that the randomized election_timeout
        will always be suit in [min_election_tick, max_election_tick).
        If it is 0, then election_tick will be chosen.

        :param max_election_tick: If it is 0, then 2 * election_tick will be chosen.

        :param read_only_option: Choose the linearizability mode or the lease mode to read data. If you don’t care about the read consistency and want a higher read performance, you can use the lease mode.
        Setting this to `LeaseBased` requires `check_quorum = true`.

        :param skip_bcast_commit: Don't broadcast an empty raft entry to notify follower to commit an entry.
        This may make follower wait a longer time to apply an entry. This configuration
        May affect proposal forwarding and follower read.

        :param batch_append: Batches every append msg if any append msg already exists

        :param priority: The election priority of this node.

        :param max_uncommitted_size: Specify maximum of uncommitted entry size.
        When this limit is reached, all proposals to append new log will be dropped

        :param max_committed_size_per_ready: Max size for committed entries in a `Ready`.
        """

class Config:
    """ """
    def __init__(
        self,
        *,
        raft_config: Optional[RaftConfig] = None,
        log_dir: Optional[str] = None,
        save_compacted_logs: Optional[bool] = None,
        compacted_log_dir: Optional[str] = None,
        compacted_log_size_threshold: Optional[int] = None,
        snapshot_interval: Optional[float] = None,
        tick_interval: Optional[float] = None,
        lmdb_map_size: Optional[int] = None,
        cluster_id: Optional[int] = None,
        terminate_on_remove: Optional[bool] = None,
        conf_change_request_timeout: Optional[float] = None,
    ) -> None:
        """ """

class RaftServiceClient:
    """ """

    @staticmethod
    async def build(self, addr: str) -> "RaftServiceClient":
        """ """
    def prepare_change_config(self, conf_change: "ConfChangeV2") -> None:
        """ """
    async def change_config(self) -> None:
        """ """
    def prepare_message(self, message: "Message") -> None:
        """ """
    async def send_message(self) -> None:
        """ """
    def prepare_propose(self, proposal: bytes) -> None:
        """ """
    async def propose(self) -> None:
        """ """

def set_snapshot_data_deserializer(cb: Callable[[bytes], str | bytes | None]) -> None:
    """ """

def set_message_context_deserializer(cb: Callable[[bytes], str | bytes | None]) -> None:
    """ """

def set_confchange_context_deserializer(
    cb: Callable[[bytes], str | bytes | None]
) -> None:
    """ """

def set_confchangev2_context_deserializer(
    cb: Callable[[bytes], str | bytes | None]
) -> None:
    """ """

def set_entry_data_deserializer(cb: Callable[[bytes], str | bytes | None]) -> None:
    """ """

def set_entry_context_deserializer(cb: Callable[[bytes], str | bytes | None]) -> None:
    """ """

def set_fsm_deserializer(cb: Callable[[bytes], str | bytes | None]) -> None:
    """ """
    ...

def set_log_entry_deserializer(cb: Callable[[bytes], str | bytes | None]) -> None:
    """ """
    ...

class ConfChangeTransition:
    """ """

    Auto: Final[Any]
    """
    Automatically use the simple protocol if possible, otherwise fall back
    to ConfChangeType::Implicit. Most applications will want to use this.
    """

    Implicit: Final[Any]
    """
    Use joint consensus unconditionally, and transition out of them
    automatically (by proposing a zero configuration change).

    This option is suitable for applications that want to minimize the time
    spent in the joint configuration and do not store the joint configuration
    in the state machine (outside of InitialState).
    """

    Explicit: Final[Any]
    """
    Use joint consensus and remain in the joint configuration until the
    application proposes a no-op configuration change. This is suitable for
    applications that want to explicitly control the transitions, for example
    to use a custom payload (via the Context field).
    """

    @staticmethod
    def from_int(v: int) -> "ConfChangeTransition": ...
    def __int__(self) -> int: ...

class MessageType:
    """ """

    MsgHup: Final[Any]
    MsgBeat: Final[Any]
    MsgPropose: Final[Any]
    MsgAppend: Final[Any]
    MsgAppendResponse: Final[Any]
    MsgRequestVote: Final[Any]
    MsgRequestVoteResponse: Final[Any]
    MsgSnapshot: Final[Any]
    MsgHeartbeat: Final[Any]
    MsgHeartbeatResponse: Final[Any]
    MsgUnreachable: Final[Any]
    MsgSnapStatus: Final[Any]
    MsgCheckQuorum: Final[Any]
    MsgTransferLeader: Final[Any]
    MsgTimeoutNow: Final[Any]
    MsgReadIndex: Final[Any]
    MsgReadIndexResp: Final[Any]
    MsgRequestPreVote: Final[Any]
    MsgRequestPreVoteResponse: Final[Any]
    @staticmethod
    def from_int(v: int) -> "MessageType": ...
    def __int__(self) -> int: ...

class ConfChangeType:
    """ """

    AddNode: Final[Any]
    AddLearnerNode: Final[Any]
    RemoveNode: Final[Any]
    @staticmethod
    def from_int(v: int) -> "ConfChangeType": ...
    def __int__(self) -> int: ...

class EntryType:
    """ """

    EntryConfChange: Final[Any]
    EntryConfChangeV2: Final[Any]
    EntryNormal: Final[Any]
    @staticmethod
    def from_int(v: int) -> "EntryType": ...
    def __int__(self) -> int: ...

class Entry:
    def get_context(self) -> bytes:
        """ """
    def set_context(self, context: bytes) -> None:
        """ """
    def get_data(self) -> bytes:
        """ """
    def set_data(self, data: bytes) -> None:
        """ """
    def get_entry_type(self) -> "EntryType":
        """ """
    def set_entry_type(self, typ: "EntryType") -> None:
        """ """
    def get_term(self) -> int:
        """ """
    def set_term(self, term: int) -> None:
        """ """
    def get_index(self) -> int:
        """ """
    def set_index(self, index: int) -> None:
        """ """
    def get_sync_log(self) -> bool:
        """
        Deprecated! It is kept for backward compatibility.
        TODO: remove it in the next major release.
        """
    def set_sync_log(self, sync_log: bool) -> None:
        """
        Deprecated! It is kept for backward compatibility.
        TODO: remove it in the next major release.
        """

class ConfChangeSingle:
    def get_node_id(self) -> int:
        """ """
    def set_node_id(self, node_id: int):
        """ """
    def get_change_type(self) -> "ConfChangeType":
        """ """
    def set_change_type(self, typ: "ConfChangeType") -> None:
        """ """

class ConfChangeV2:
    def get_changes(self) -> list["ConfChangeSingle"]:
        """ """
    def set_changes(
        self, changes: list["ConfChangeSingle"] | list["ConfChangeSingle"]
    ) -> None:
        """ """
    def get_context(self) -> bytes:
        """ """
    def set_context(self, context: bytes) -> None:
        """ """
    def get_transition(self) -> "ConfChangeTransition":
        """ """
    def set_transition(self, transition: "ConfChangeTransition") -> None:
        """ """
    # def enter_joint(self) -> Optional[bool]:
    #     """
    #     Checks if uses Joint Consensus.

    #     It will return Some if and only if this config change will use Joint Consensus,
    #     which is the case if it contains more than one change or if the use of Joint
    #     Consensus was requested explicitly. The bool indicates whether the Joint State
    #     will be left automatically.
    #     """

    # def leave_joint(self) -> bool:
    #     """
    #     Checks if the configuration change leaves a joint configuration.

    #     This is the case if the ConfChangeV2 is zero, with the possible exception of
    #     the Context field.
    #     """

class Message:
    def get_commit(self) -> int:
        """ """
    def set_commit(self, commit: int) -> None:
        """ """
    def get_commit_term(self) -> int:
        """ """
    def set_commit_term(self, commit_term: int) -> None:
        """ """
    def get_from(self) -> int:
        """ """
    def set_from(self, from_: int) -> None:
        """ """
    def get_index(self) -> int:
        """ """
    def set_index(self, index: int) -> None:
        """ """
    def get_term(self) -> int:
        """ """
    def set_term(self, term: int) -> None:
        """ """
    def get_log_term(self) -> int:
        """ """
    def set_log_term(self, log_index: int) -> None:
        """ """
    def get_priority(self) -> int:
        """ """
    def set_priority(self, priority: int) -> None:
        """ """
    def get_context(self) -> bytes:
        """ """
    def set_context(self, context: bytes) -> None:
        """ """
    def get_reject_hint(self) -> int:
        """ """
    def set_reject_hint(self, reject_hint: int) -> None:
        """ """
    def get_entries(self) -> list["Entry"]:
        """ """
    def set_entries(self, ents: list["Entry"]) -> None:
        """ """
    def get_msg_type(self) -> "MessageType":
        """ """
    def set_msg_type(self, typ: "MessageType") -> None:
        """ """
    def get_reject(self) -> bool:
        """ """
    def set_reject(self, reject: bool) -> None:
        """ """
    # def get_snapshot(self) -> "SnapshotRef":
    #     """ """
    # def set_snapshot(self, snapshot: "Snapshot" | "SnapshotRef") -> None:
    #     """ """
    def get_to(self) -> int:
        """ """
    def set_to(self, to: int) -> None:
        """ """
    def get_request_snapshot(self) -> int:
        """ """
    def set_request_snapshot(self, request_snapshot: int) -> None:
        """ """
    def has_snapshot(self) -> bool:
        """ """
    def get_deprecated_priority(self) -> int:
        """ """
    def set_deprecated_priority(self, v: int) -> None:
        """ """

async def cli_main(argv: list[str]) -> None:
    """ """
    ...



import asyncio
import os
from harness.state_machine import HashStore
from raftify import Config, Peers, Raft, RaftConfig, Logger


RAFTS: dict[int, Raft] = {}


def build_config() -> Config:
    raft_cfg = RaftConfig()
    cfg = Config(
        raft_cfg,
        log_dir="./logs",
        compacted_log_dir="./logs",
    )

    return cfg


async def wait_for_termination():
    for raft in RAFTS.values():
        while not raft.is_finished():
            await asyncio.sleep(1.0)


async def run_raft(node_id: int, peers: Peers):
    peer = peers.get(node_id)
    cfg = build_config()

    store = HashStore()
    logger = Logger.default()
    raft = Raft.build(node_id, peer, store, cfg, logger, peers)
    await raft.run()

    RAFTS[node_id] = raft


async def run_rafts(peers: Peers):
    # TODO: This should be configurable
    os.environ["RUST_LOG"] = "debug"

    for node_id, _ in peers.items():
        await run_raft(node_id, peers)


async def handle_bootstrap(peers: Peers):
    leader_addr = peers.get(1)

    for (node_id, _) in peers.items():
        if node_id != 1:
            raft = RAFTS.get(node_id)
            raft.prepare_member_bootstrap_ready(leader_addr, node_id)
            await raft.member_bootstrap_ready()



import os
from pathlib import Path

from setuptools import setup

ROOT_PATH = os.path.abspath(os.path.dirname(__file__))
VERSION = (Path(__file__).parent / "VERSION").read_text().strip()
README = (Path(__file__).parent / "README.md").read_text().strip()


def get_requirements(env: str = None):
    requirements = "requirements"
    if env is not None:
        requirements = f"{requirements}-{env}"
    with open(f"{requirements}.txt".format(env)) as fp:
        return [x.strip() for x in fp.read().split("\n") if not x.startswith("#")]


default_requires = get_requirements()

setup(
    name="raftify_cli",
    version=VERSION,
    description="Experimental Async Raft framework for Python application",
    long_description=README,
    long_description_content_type='text/markdown',
    author="Lablup Inc.",
    maintainer="jopemachine",
    maintainer_email="jopemachine@naver.com",
    url="https://github.com/lablup/raftify",
    license="Apache License 2.0",
    package_dir={"raftify_cli": "raftify_cli"},
    package_data={"": ["VERSION"]},
    python_requires=">=3.10",
    install_requires=default_requires,
    zip_safe=False,
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'raftify_cli = raftify_cli.cli:main'
        ]
    }
)



import argparse
import asyncio
import pickle
from contextlib import suppress
from pathlib import Path
from typing import Any, Iterable, Optional

import tomli
from aiohttp import web
from aiohttp.web import AbstractRouteDef, RouteTableDef
from raftify import (
    Config,
    Peers,
    Raft,
    RaftConfig,
    Logger,
    OverflowStrategy,
    set_confchange_context_deserializer,
    set_confchangev2_context_deserializer,
    set_entry_context_deserializer,
    set_entry_data_deserializer,
    set_message_context_deserializer,
    set_snapshot_data_deserializer,
)


def load_peers() -> Peers:
    path = Path(__file__).parent / "cluster_config.toml"
    cfg = tomli.loads(path.read_text())["raft"]["peers"]

    return Peers(
        {int(entry["node_id"]): f"{entry['host']}:{entry['port']}" for entry in cfg}
    )


def build_config() -> Config:
    raft_cfg = RaftConfig()
    cfg = Config(
        raft_cfg,
        log_dir="./logs",
        compacted_log_dir="./logs",
    )

    return cfg


class WebServer:
    def __init__(
        self, addr: str, routes: Iterable[AbstractRouteDef], state: dict[str, Any]
    ):
        self.app = web.Application()
        self.app.add_routes(routes)
        self.app["state"] = state
        self.host, self.port = addr.split(":")
        self.runner = None

    async def __aenter__(self):
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, self.host, self.port)
        await self.site.start()
        return self.runner

    async def __aexit__(self, exc_type, exc, tb):
        await self.runner.cleanup()
        await self.runner.shutdown()


routes = RouteTableDef()
"""
APIs of the web servers to interact with the RaftServers.
"""


@routes.get("/get/{id}")
async def get(request: web.Request) -> web.Response:
    store: HashStore = request.app["state"]["store"]
    id = request.match_info["id"]
    return web.Response(text=store.get(id))


@routes.get("/leader")
async def leader(request: web.Request) -> web.Response:
    raft: Raft = request.app["state"]["raft"]
    leader_id = str(await raft.get_raft_node().get_leader_id())
    return web.Response(text=leader_id)


@routes.get("/size")
async def size(request: web.Request) -> web.Response:
    raft: Raft = request.app["state"]["raft"]
    size = str(await raft.get_raft_node().get_cluster_size())
    return web.Response(text=size)


@routes.get("/put/{id}/{value}")
async def put(request: web.Request) -> web.Response:
    raft: Raft = request.app["state"]["raft"]
    id, value = request.match_info["id"], request.match_info["value"]
    message = SetCommand(id, value)

    raft_node = raft.get_raft_node()
    raft_node.prepare_proposal(message.encode())
    await raft_node.propose()
    return web.Response(text="OK")


class SetCommand:
    """
    Represent simple key-value command.
    Use pickle to serialize the data.
    """

    def __init__(self, key: str, value: str) -> None:
        self.key = key
        self.value = value

    def encode(self) -> bytes:
        return pickle.dumps(self.__dict__)

    @classmethod
    def decode(cls, packed: bytes) -> "SetCommand":
        unpacked = pickle.loads(packed)
        return cls(unpacked["key"], unpacked["value"])


def pickle_deserialize(data: bytes) -> str | None:
    if data == b"":
        return None

    if pickle.PROTO in data:
        r = pickle.loads(data[data.index(pickle.PROTO) :])
        return r

    # Not pickle data
    return None


def register_custom_deserializer() -> None:
    """
    Initialize the custom deserializers.
    """

    set_confchange_context_deserializer(pickle_deserialize)
    set_confchangev2_context_deserializer(pickle_deserialize)
    set_entry_context_deserializer(pickle_deserialize)
    set_entry_data_deserializer(pickle_deserialize)
    set_message_context_deserializer(pickle_deserialize)
    set_snapshot_data_deserializer(pickle_deserialize)


class HashStore:
    """
    A simple key-value store that stores data in memory.
    Use pickle to serialize the data.
    """

    def __init__(self):
        self._store = dict()

    def get(self, key: str) -> Optional[str]:
        return self._store.get(key)

    def as_dict(self) -> dict:
        return self._store

    def apply(self, msg: bytes) -> bytes:
        message = SetCommand.decode(msg)
        self._store[message.key] = message.value
        return msg

    def snapshot(self) -> bytes:
        return pickle.dumps(self._store)

    def restore(self, snapshot: bytes) -> None:
        self._store = pickle.loads(snapshot)


async def wait_for_termination(raft: Raft):
    while True:
        if raft.is_finished():
            break
        await asyncio.sleep(1)


async def main():
    register_custom_deserializer()
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--ignore-static-bootstrap", action=argparse.BooleanOptionalAction, default=None
    )
    parser.add_argument("--raft-addr", default=None)
    parser.add_argument("--peer-addr", default=None)
    parser.add_argument("--web-server", default=None)
    args = parser.parse_args()

    raft_addr = args.raft_addr
    peer_addr = args.peer_addr
    web_server_addr = args.web_server
    ignore_static_bootstrap = args.ignore_static_bootstrap

    peers = load_peers() if not ignore_static_bootstrap else None

    cfg = build_config()
    logger = Logger.default()
    store = HashStore()
    tasks = []

    if peer_addr:
        if not peers:
            join_ticket = await Raft.request_id(peer_addr)
            node_id = join_ticket.get_reserved_id()
        else:
            node_id = peers.get_node_id_by_addr(raft_addr)

        raft = Raft.build(node_id, raft_addr, store, cfg, logger, peers)
        await raft.run()

        if not peers:
            raft.prepare_join(join_ticket)
            await raft.join()
        else:
            leader_addr = peers.get(1)
            raft.prepare_member_bootstrap_ready(leader_addr, node_id)
            await raft.member_bootstrap_ready()
    else:
        raft = Raft.build(1, raft_addr, store, cfg, logger, peers)
        await raft.run()

    tasks.append(asyncio.create_task(wait_for_termination(raft)))

    async with WebServer(web_server_addr, routes, {"raft": raft, "store": store}):
        await asyncio.gather(*tasks)


if __name__ == "__main__":
    with suppress(KeyboardInterrupt):
        asyncio.run(main())
